@@ -52,6 +52,7 @@ import           What4.Protocol.Online
 import           What4.Protocol.SMTLib2
 import           What4.SatResult
 import           What4.Solver.Adapter
+import qualified What4.Solver.Bitwuzla as Bitwuzla
 import qualified What4.Solver.CVC4 as CVC4
 import qualified What4.Solver.CVC5 as CVC5
 import qualified What4.Solver.Z3 as Z3
@@ -141,6 +142,18 @@ withCVC5 action = withSym FloatRealRepr $ \sym -> do
     (\(h,s) -> void $ try @SomeException (shutdownSolverProcess s `finally` maybeClose h))
     (\(_,s) -> action sym s)
 
+withBitwuzla
+  :: (forall t . SimpleExprBuilder t (Flags FloatReal) -> SolverProcess t (Writer Bitwuzla.Bitwuzla) -> IO a)
+  -> IO a
+withBitwuzla action = withSym FloatRealRepr $ \sym -> do
+  extendConfig Bitwuzla.bitwuzlaOptions (getConfiguration sym)
+  bracket
+    (do h <- if debugOutputFiles then Just <$> openFile "bitwuzla.out" WriteMode else return Nothing
+        s <- startSolverProcess (defaultFeatures Bitwuzla.Bitwuzla) h sym
+        return (h,s))
+    (\(h,s) -> void $ try @SomeException (shutdownSolverProcess s `finally` maybeClose h))
+    (\(_,s) -> action sym s)
+
 withModel
   :: Session t Z3.Z3
   -> BoolExpr t
@@ -1114,6 +1127,40 @@ issue315Test sym solver = do
 
       _ -> fail "expected satisfible model"
 
+-- | A regression test for #329.
+issue329Test ::
+  OnlineSolver solver =>
+  SimpleExprBuilder t fs ->
+  SolverProcess t solver ->
+  IO ()
+issue329Test sym solver = do
+    -- This test shows that the following proposition (written in Cryptol) is
+    -- falsifiable when `x == 2`:
+    --
+    --   \(x : [2]) -> x == 2 ==> ((toSignedInteger x - 1) == -1)
+    let w = knownNat @2
+    x <- freshConstant sym (safeSymbol "x") (BaseBVRepr w)
+    bvTwo <- bvLit sym w (BV.mkBV w 2)
+    xEqBvTwo <- bvEq sym x bvTwo
+    notXEqBvTwo <- notPred sym xEqBvTwo
+    xsi <- sbvToInteger sym x
+    intOne <- intLit sym 1
+    intNegOne <- intLit sym (-1)
+    xsiDec <- intSub sym xsi intOne
+    xsiDecEqIntNegOne <- intEq sym xsiDec intNegOne
+    p <- orPred sym notXEqBvTwo xsiDecEqIntNegOne
+
+    -- Check if `p` is not valid for all `x` by checking if its negation
+    -- (`notP`) is satisfiable. If `notP` is satisfiable, then we have found a
+    -- counterexample to `p`.
+    notP <- notPred sym p
+    checkSatisfiableWithModel solver "test" notP $ \case
+      Sat fn ->
+        do xEval <- groundEval fn x
+           (xEval == BV.mkBV w 2) @? "result other than 2"
+
+      _ -> fail "expected satisfible model"
+
 -- | These tests simply ensure that no exceptions are raised.
 testSolverInfo :: TestTree
 testSolverInfo = testGroup "solver info queries" $
@@ -1226,7 +1273,7 @@ testResolveSymBV searchStrat =
 main :: IO ()
 main = do
   testLevel <- TestLevel . fromMaybe "0" <$> lookupEnv "CI_TEST_LEVEL"
-  let solverNames = SolverName <$> [ "cvc4", "cvc5", "yices", "z3" ]
+  let solverNames = SolverName <$> [ "bitwuzla", "cvc4", "cvc5", "yices", "z3" ]
   solvers <- reportSolverVersions testLevel id
              =<< (zip solverNames <$> mapM getSolverVersion solverNames)
   let z3Tests =
@@ -1277,6 +1324,7 @@ main = do
 
         , testCase "Z3 #182 test case" $ withOnlineZ3 issue182Test
         , testCase "Z3 #315 test case" $ withOnlineZ3 issue315Test
+        , testCase "Z3 #329 test case" $ withOnlineZ3 issue329Test
 
         , arrayCopyTest
         , arraySetTest
@@ -1328,6 +1376,7 @@ main = do
 
         , cvcTestCase "#182 test case" $ withCVC issue182Test
         , cvcTestCase "#315 test case" $ withCVC issue315Test
+        , cvcTestCase "#329 test case" $ withCVC issue329Test
         ]
   let cvc4Tests = cvcTests CVC4
   let cvc5Tests = cvcTests CVC5
@@ -1342,6 +1391,10 @@ main = do
         , testCase "Yices rounding" $ withYices roundingTest
         , testCase "Yices #182 test case" $ withYices issue182Test
         , testCase "Yices #315 test case" $ withYices issue315Test
+        , testCase "Yices #329 test case" $ withYices issue329Test
+        ]
+  let bitwuzlaTests =
+        [ testCase "Bitwuzla multidim array" $ withBitwuzla multidimArrayTest
         ]
   let skipIfNotPresent nm = if SolverName nm `elem` (fst <$> solvers) then id
                             else fmap (ignoreTestBecause (nm <> " not present"))
@@ -1362,6 +1415,7 @@ main = do
     , testUnsafeSetAbstractValue1
     , testUnsafeSetAbstractValue2
     ]
+    <> (skipIfNotPresent "bitwuzla" bitwuzlaTests)
     <> (skipIfNotPresent "cvc4" cvc4Tests)
     <> (skipIfNotPresent "cvc5" cvc5Tests)
     <> (skipIfNotPresent "yices" yicesTests)
